local M = {}

local debug = require("../debug")
local mathutil = require("../mathutil")

M.type = "auxilliary"
M.relevantDevice = nil
M.defaultOrder = 900

local constants = {
    -- From shiftLogic
    rpmToAV = 0.104719755,
    avToRPM = 9.549296596425384,
    -- Regen braking simulation
    maxRegenPercent = 0.85,
    minRegenSpeed = 4,
    maxRegenSpeed = 22,
    regenThrottleMidpoint = 0.25,
    -- Engine accel delay
    minBatteryLevel = 0.3,
    maxBatterylevel = 0.9,
    maxEngineShiftSpeed = 45,
    minEngineEngageSpeed = -10,
    maxEngineEngageSpeed = -6,
    minThrusterFadeSpeed = -10,
    maxThrusterFadeSpeed = 0,
    -- Engine power generation
    minEnginePowerGenRpm = 300,
    maxEnginePowerGenRpm = 1600,
    maxEnginePowerGenKw = 80,
    -- Smoothing
    sThrottleIn = 1.65,
    sThrottleOut = 1.85,
    -- Sounds
    inverterBaseVolume = 0.85,
    inverterPitchMinSpeed = 1,
    inverterPitchMaxSpeed = 3
}

local smoothers = {
    throttle = nil
}

local sounds = {
    inverterWhine = nil
}

local gearbox = nil
local engine = nil
local battery = nil

local enableAutoStop = false
local throttle = 0
local effectiveThrottle = 0
local wheelMph = 0
local gearIndex = 0
local gear = nil
local parkingBrake = false
local ignition = false
local tractionEffort = 0
local regenEffort = 0
local accelEffort = 0
local lastAutoStop = false
local lastIgnition = nil
local hasMovedSinceEngineStart = false
local timeSinceEngineStart = 0
local initialGearboxShiftTime = 0

local function updateValues(dt)
    throttle = smoothers.throttle:get(input.throttle or 0.0, dt)
    wheelMph = electrics.values.wheelspeed * 2.236
    gearIndex = electrics.values.gearIndex
    gear = electrics.values.gear
    parkingBrake = electrics.values.parkingbrake > 0
    ignition = electrics.values.ignition

    -- Fix wheelspeed being above zero when vehicle is actually stopped
    if wheelMph < 0.5 and (parkingBrake or input.brake > 0.1) then
        electrics.values.wheelspeed = 0.0
    end
end

local function updateRegen(dt)
    -- Calculate regenerative braking (simulated with negative thrust)

    local speedCoef = (wheelMph - constants.minRegenSpeed) / (constants.maxRegenSpeed - constants.minRegenSpeed)
    local regenCoef = mathutil.clamp(speedCoef, 0.0, 1.0)
    local actualMidpoint = mathutil.clamp(constants.regenThrottleMidpoint * regenCoef, 0.01, 1.0)
    local pedalRegenAmount = mathutil.clamp(actualMidpoint - throttle, 0.0, actualMidpoint) / actualMidpoint
    local actualThrottleAmount = mathutil.clamp((throttle - actualMidpoint) / (1.0 - actualMidpoint), 0.0, 1.0)

    if gearIndex > 0 then
        effectiveThrottle = actualThrottleAmount
        regenEffort = regenCoef * pedalRegenAmount
    else
        effectiveThrottle = throttle
        regenEffort = 0.0
    end

    debug.dText(0, "Throttle: " .. (throttle * 100) .. "%")
end

local function updateAccel(dt)
    -- Engine doesn't kick in right away; some torque comes from electric at first
    -- We use a thruster to simulate the electric part of the drivetrain

    local batteryLevel = battery.getChargeLevel()
    local batteryUsableLevel =
        mathutil.clamp(
        (batteryLevel - constants.minBatteryLevel) / (constants.maxBatterylevel - constants.minBatteryLevel),
        0.0,
        1.0
    )
    local engineDelayShift = batteryUsableLevel * constants.maxEngineShiftSpeed
    local minEngineEngageSpeed = constants.minEngineEngageSpeed + engineDelayShift
    local maxEngineEngageSpeed = constants.maxEngineEngageSpeed + engineDelayShift
    local minThrusterFadeSpeed = constants.minThrusterFadeSpeed + engineDelayShift
    local maxThrusterFadeSpeed = constants.maxThrusterFadeSpeed + engineDelayShift
    local speedCoefEng = (wheelMph - minEngineEngageSpeed) / (maxEngineEngageSpeed - minEngineEngageSpeed)
    local speedCoefThr = (wheelMph - minThrusterFadeSpeed) / (maxThrusterFadeSpeed - minThrusterFadeSpeed)
    local engineDelayCoef = mathutil.clamp(mathutil.exp(speedCoefEng, 1 / 9, 10, -1 / 9), 0.1, 1.0)
    local maxThrusterPower = clamp( batteryLevel * 2, 0.0, 1.0 )
    local thrusterCoef = throttle * maxThrusterPower * (1.0 - mathutil.clamp(speedCoefThr, 0.0, 1.0))

    if gearIndex > 0 then
        electrics.values.throttleOverride = effectiveThrottle * engineDelayCoef
        accelEffort = thrusterCoef
        tractionEffort = thrusterCoef
    elseif gear == "R" then
        local reverseCoef = throttle * (1.0 - mathutil.clamp(wheelMph / 12, 0.0, 1.0))

        electrics.values.throttleOverride = 0.0 -- No throttle in reverse
        accelEffort = -reverseCoef
        tractionEffort = reverseCoef
    else
        electrics.values.throttleOverride = effectiveThrottle
        accelEffort = 0.0
        tractionEffort = 0.0
    end

    debug.dText(1, "Engine delay %: " .. engineDelayCoef * 100 .. "%")
    debug.dText(2, "Thruster %: " .. thrusterCoef * 100 .. "%")
end

local function updateThruster(dt)
    -- Regen is positive values for negative thrust, so negate it and add it to accel
    local trueRegenEffort = -(regenEffort * constants.maxRegenPercent)
    local thrusterEffort = accelEffort + trueRegenEffort

    electrics.values.hybrid_thruster = thrusterEffort
end

local function updatePowerConsumption(dt)
    local thrusterEffort = electrics.values.hybrid_thruster
    local tractionPowerConsumption = thrusterEffort * M.motorPower
    local minGenRpm = constants.minEnginePowerGenRpm
    local maxGenRpm = constants.maxEnginePowerGenRpm
    local rpm = electrics.values.rpm
    local enginePowerGen = constants.maxEnginePowerGenKw * clamp((rpm - minGenRpm) / (maxGenRpm - minGenRpm), 0.0, 1.0)

    M.powerConsumption = tractionPowerConsumption - enginePowerGen
end

local function updateGearbox(dt)
    local engineOn = engine.ignitionCoef > 0 and engine.starterEngagedCoef == 0

    if engineOn then
        timeSinceEngineStart = math.min(timeSinceEngineStart + dt, 10.0)
    else
        timeSinceEngineStart = 0
    end

    if gearbox.type == "cvtGearbox" then
        gearbox.lockCoef = math.min(timeSinceEngineStart / 0.85, 1.0)
    elseif gearbox.type == "automaticGearbox" then
        if gearIndex > 0 then
            if engineOn and gearbox.mode == "neutral" then
                gearbox:setMode("drive")
                gearbox:setGearIndex(math.max(wheelMph / 6, 1)) -- very rough estimate of gear we should be in
            elseif not engineOn and gearbox.mode ~= "neutral" then
                gearbox:setMode("neutral")
            end
        end
    end

    electrics.values.engineThrottle = math.min(electrics.values.engineThrottle, math.min(timeSinceEngineStart, 1.0))
end

local function updateEngine(dt)
    local autoStop =
        enableAutoStop and hasMovedSinceEngineStart and (gearIndex > 0 or gear == "N") and
        (input.brake > 0.15 or parkingBrake) and
        wheelMph < 1

    if not autoStop and wheelMph > 4 then
        -- Allow engine to shut down again

        hasMovedSinceEngineStart = true
    end

    -- Obey actual game ignition control

    if lastIgnition == nil then
        lastIgnition = ignition
    end

    if ignition ~= lastIgnition then
        autoStop = not ignition
        hasMovedSinceEngineStart = false
    end

    -- Start or stop the engine if needed

    if not autoStop and lastAutoStop then
        engine:activateStarter()
        engine:setIgnition(ignition and 1 or 0)
        hasMovedSinceEngineStart = false
    elseif autoStop then
        engine:setIgnition(0)
        engine:deactivateStarter()
    end

    lastAutoStop = autoStop
    lastIgnition = ignition
end

local function updateSounds()
    -- Update sounds
    local inverterPitchCoef =
        (wheelMph - constants.inverterPitchMinSpeed) /
        (constants.inverterPitchMaxSpeed - constants.inverterPitchMinSpeed)
    local inverterEffort = math.max(tractionEffort, regenEffort)
    local volume = constants.inverterBaseVolume * mathutil.clamp(inverterEffort, 0.0, 1.0)
    local pitch = mathutil.clamp(inverterPitchCoef, 0.4, 1.0)

    obj:setVolumePitch(sounds.inverterWhine, volume, pitch)
end

local function initSounds()
    sounds.inverterWhine =
        obj:createSFXSource(
        "vehicles/citybus/sounds/inverter.ogg",
        "AudioDefaultLoop3D",
        "inverterWhine",
        engine.engineNodeID
    )
end

local function init(jbeamData)
    M.motorPower = jbeamData.motorPower or 190.0 -- kW
    M.powerConsumption = 0.0

    smoothers.throttle = newTemporalSmoothingNonLinear(constants.sThrottleIn, constants.sThrottleOut)
end

local function initSecondStage()
    gearbox = powertrain.getDevice("gearbox")
    engine = powertrain.getDevice("mainEngine")
    battery = controller.getController("battery")

    -- TODO
    bat = energyStorage.getStorage("mainBattery")
    bat.storedEnergy = 0.8 * bat.initialStoredEnergy

    if gearbox then
        initialGearboxFriction = gearbox.friction

        if gearbox.type == "automaticGearbox" then
            initialGearboxShiftTime = gearbox.minimumGearChangeTime
        end
    end

    if battery then
        battery.registerHybridSystem(M)
    end
end

local function updateGFX(dt)
    updateValues(dt)
    updateRegen(dt)
    updateAccel(dt)
    updateThruster(dt)
    updatePowerConsumption(dt)

    if engine then
        if gearbox then
            updateGearbox(dt)
        end

        updateEngine(dt)
    end

    updateSounds()
end

M.init = init
M.initSounds = initSounds
M.initSecondStage = initSecondStage
M.updateGFX = updateGFX

return M
